# -*- coding: utf-8 -*-
"""
Created on Tue Mar 24 07:35:11 2020

@author: Marie-Laure Besson
Traduction du programme "star pusher de Al Sweigart"
"""

import random, copy, os, pygame

FPS = 30 
LARGEUR_ECRAN = 800 
HAUTEUR_ECRAN = 600
MILIEU_LARGEUR_ECRAN = int(LARGEUR_ECRAN / 2)
MILIEU_HAUTEUR_ECRAN = int(HAUTEUR_ECRAN / 2)

LARGEUR_TUILE = 50
HAUTEUR_TUILE = 85
SOL_TUILE = 40

PCT_DECO = 20

BLEUCLAIR = (  0, 170, 255)
BLANC      = (255, 255, 255)

# =============================================================================
# Fonctions de création d'un niveau de jeu
# =============================================================================
def lectureFichierNiveaux(nomFichier):
    assert os.path.exists(nomFichier), 'Je ne trouve pas le fichier de niveaux: %s' % (nomFichier)
    fichierCarte = open(nomFichier, 'r')
    # Chaque niveau se termine par une ligne vide
    contenu = fichierCarte.readlines()+ ['\r\n']
    fichierCarte.close()
    #liste de tous les niveaux
    niveaux = []
    numNiveau = 0
    #liste contenant les lignes de texte d'un niveau
    lignesTexteCarte = [] 
    #la carte créée à partir des données de lignesTexteCarte
    carte = [] 
    for ligneNum in range(len(contenu)):
        #on traite chaque ligne du niveau
        ligne = contenu[ligneNum].rstrip('\r\n')

        if ';' in ligne:
            #une ligne démarrant avec ; est une ligne de commentaire
            #que l'on doit ignorer
            ligne = ligne[:ligne.find(';')]

        if ligne != '':
            # une ligne qui n'est pas vide, appartient à la carte de ce niveau.
            lignesTexteCarte.append(ligne)
        elif ligne == '' and len(lignesTexteCarte) > 0:
            # Une ligne vide, indique la fin de la carte d'un niveau.
            #conversion du texte contenu dans lignesTexteCarte
            #en une carte de niveau
           
            # Trouver la ligne la plus longue de la carte.
            largeurMax = -1
            for i in range(len(lignesTexteCarte)):
                if len(lignesTexteCarte[i]) > largeurMax:
                    largeurMax = len(lignesTexteCarte[i])
            # Ajouter des espaces aux lignes les plus courtes de façon à ce que la carte 
            #soit rectangulaire.
            for i in range(len(lignesTexteCarte)):
                lignesTexteCarte[i] += ' ' * (largeurMax - len(lignesTexteCarte[i]))

            # Convertir lignesTexteCarte en une carte.
            for x in range(len(lignesTexteCarte[0])):
                carte.append([])
            for y in range(len(lignesTexteCarte)):
                for x in range(largeurMax):
                    carte[x].append(lignesTexteCarte[y][x])

            #boucle dans la carte pour trouver les caractères @, . et $
            #indiquant l'état de départ du jeu
            #position de départ du joueur
            startx = None 
            starty = None
            #liste de chaque coordonnées x, y des objectifs
            objectifs = [] 
            #liste de chaque coordonnées x, y des étoiles
            etoiles = [] 
            for x in range(largeurMax):
                for y in range(len(carte[x])):
                    if carte[x][y] in ('@', '+'):
                        #'@' position départ joueur
                        #'+' un objectif avec le joueur posé dessus au départ
                        startx = x
                        starty = y
                    if carte[x][y] in ('.', '+', '*'):
                        # '.' objectif
                        # '*' un objectif avec une étoile posée
                        # '+' un objectif avec le joueur posé dessus au départ
                        objectifs.append((x, y))
                    if carte[x][y] in ('$', '*'):
                        # '$' étoile
                        # '*' un objectif avec une étoile posée
                        etoiles.append((x, y))

            # Tests élémentaires:
            assert startx != None and starty != None, 'Dans le niveau %s (autour de la ligne %s) dans %s il manque un "@" ou un "+" pour marquer le point de départ du joueur.' % (numNiveau+1, ligneNum, nomFichier)
            assert len(objectifs) > 0, 'Dans le niveau %s (autour de la ligne %s) dans %s il doit y avoir au moins un objectif.' % (numNiveau+1, ligneNum, nomFichier)
            assert len(etoiles) >= len(objectifs), 'Dans le niveau %s (autour de la ligne %s) dans %s il y a %s objectifs et seulement %s étoiles. Ce niveau ne peut pas être résolu' % (numNiveau+1, ligneNum, nomFichier, len(objectifs), len(etoiles))

            # Etat du jeu au départ du niveau
            jeuEtat = {'joueur': (startx, starty),                            
                      'compteur': 0,
                      'etoiles': etoiles}
            niveau = {'largeur': largeurMax,
                        'hauteur': len(carte),
                        'carte': carte,
                        'objectifs': objectifs,
                        'etatInitial': jeuEtat}
            niveaux.append(niveau)

            # Remise à 0 des variables pour lire la carte suivante
            lignesTexteCarte = []
            carte = []
            jeuEtat = {}
            numNiveau += 1
    return niveaux

def modifierCarte(carte, startxy):
    """Réalise une copie de la carte de départ et la modifie.
        * Les caractères '#' représentants des murs qui sont en réalité des coins, 
          sont remplacés par un caractère 'x' qui indiquera un coin.
        * On fait la dictinction entre tuile d'extérieur et tuile d'intérieur: 
          Le symbole d'une tuile d'intérieure est 'o' à la place d'un  ' '
        * Les décorations supplémentaires (arbres, rochers) sont ajoutées
          de façon aléatoire aux tuiles extérieures.

    Renvoie la carte modifiée."""

    startx, starty = startxy 

    # Copie la carte de départ, pour ne pas la modifier
    carteCopy = copy.deepcopy(carte)

    # Supprime tous les symboles qui ne correspondent pas à des murs
    for x in range(len(carteCopy)):
        for y in range(len(carteCopy[0])):
            if carteCopy[x][y] in ('$', '.', '@', '+', '*'):
                carteCopy[x][y] = ' '

    # remplissage par diffusion : remplace les caractères représentant 
    # des tuiles externes, par des caratères de tuiles internes pour toutes
    #les tuiles situées entre les murs.
    remplissageDiffusion(carteCopy, startx, starty, ' ', 'o')

    # Conversion des murs adjacents en coins.
    for x in range(len(carteCopy)):
        for y in range(len(carteCopy[0])):

            if carteCopy[x][y] == '#':
                if (estMur(carteCopy, x, y-1) and estMur(carteCopy, x+1, y)) or \
                   (estMur(carteCopy, x+1, y) and estMur(carteCopy, x, y+1)) or \
                   (estMur(carteCopy, x, y+1) and estMur(carteCopy, x-1, y)) or \
                   (estMur(carteCopy, x-1, y) and estMur(carteCopy, x, y-1)):
                    carteCopy[x][y] = 'x'
            #décoration des tuiles externes
            elif carteCopy[x][y] == ' ' and random.randint(0, 99) < PCT_DECO:
                carteCopy[x][y] = random.choice(list(LiaisonExterieur .keys()))

    return carteCopy

def estMur(carte, x, y):
    """Renvoie True si à la position (x,y) sur la carte
    on a un mur. Sinon retourne False."""
    if x < 0 or x >= len(carte) or y < 0 or y >= len(carte[x]):
        return False # x et y ne sont pas sur la carte
    elif carte[x][y] in ('#', 'x'):
        return True # en x, y il y a bien un mur
    return False

def remplissageDiffusion(carte,x,y,ancienCaractere, nouveauCaractere):
    """ Remplace à l'intérieur des murs, les caractères ' ' par un caractère 'o'
    """
    espaceATester=[]
    if carte[x][y]== ancienCaractere:
        espaceATester.append((x,y))
    
    while espaceATester !=[]:
        x,y = espaceATester.pop()
        carte[x][y]=nouveauCaractere
        #test case à droite (Est)
        if x < len(carte)-1 and carte[x+1][y]== ancienCaractere:
            espaceATester.append((x+1,y))
         #test case à gauche (Nord)    
        if x > 0 and carte[x-1][y]== ancienCaractere:
            espaceATester.append((x-1,y))
         #test case en dessous (Sud)
        if y < len(carte[x])-1 and carte[x][y+1]== ancienCaractere:
            espaceATester.append((x,y+1))
         #test case au dessus (Nord)
        if y > 0 and carte[x][y-1]== ancienCaractere:
            espaceATester.append((x,y-1))

def dessinerScene(carte, jeuEtat, objectifs):
    """Dessine la carte sur une surface en y incluant le joueur et les étoiles.
    Ici on fait donc le lien entre les caractères enregistrés sur une carte
    et les images qui leur correspondent.
    """
    # surfaceCarte est un objet Surface sur lequel les tuiles sont dessinées.
    surfaceCarteLargeur = len(carte) * LARGEUR_TUILE
    surfaceCarteHauteur = (len(carte[0]) - 1) * SOL_TUILE + HAUTEUR_TUILE
    surfaceCarte = pygame.Surface((surfaceCarteLargeur, surfaceCarteHauteur))
    surfaceCarte.fill(BLEUCLAIR) 

    # Dessin des tuiles
    for x in range(len(carte)):
        for y in range(len(carte[x])):
            spaceRect = pygame.Rect((x * LARGEUR_TUILE, y * SOL_TUILE, LARGEUR_TUILE, HAUTEUR_TUILE))
            # On dessine en premier les tuiles de base : sol/mur
            if carte[x][y] in LiaisonTuile:
                tuileBase = LiaisonTuile[carte[x][y]]
            elif carte[x][y] in LiaisonExterieur:
                tuileBase = LiaisonTuile[' ']            
            surfaceCarte.blit(tuileBase, spaceRect)
            
            #on rajoute les décorations supplémentaires extérieures
            if carte[x][y] in LiaisonExterieur:
                surfaceCarte.blit(LiaisonExterieur[carte[x][y]], spaceRect)
            #dessin des étoiles
            elif (x, y) in jeuEtat['etoiles']:
                if (x, y) in objectifs:
                    # Si un objectif et une étoile sont sur la même case
                    #on dessine l'objectif en premier.
                    surfaceCarte.blit(DicoImage['objectif couvert'], spaceRect)
                # Ensuite on dessine l'étoile.
                surfaceCarte.blit(DicoImage['etoile'], spaceRect)
            #dessine les objectifs, sans étoiles dessus
            elif (x, y) in objectifs:
                surfaceCarte.blit(DicoImage['objectif decouvert'], spaceRect)

            # Dessine le joueur
            if (x, y) == jeuEtat['joueur']:
                #la valeur contenue dans joueurActuel correspond à une clé
                #du dictionnaire ImagesJoueur et indique quelle image représente
                #actuellement le joueur
                surfaceCarte.blit(ImagesJoueur[joueurActuel], spaceRect)
    
    return surfaceCarte
# =============================================================================
# 
# =============================================================================
pygame.init()
FPSCLOCK = pygame.time.Clock()
ecran = pygame.display.set_mode((LARGEUR_ECRAN, HAUTEUR_ECRAN))
pygame.display.set_caption('Star Pusher')
FONT = pygame.font.SysFont('comicsansms',18)

#Distionnaire des images chargées
DicoImage = {'objectif decouvert': pygame.image.load('RedSelector.png'),
              'objectif couvert': pygame.image.load('Selector.png'),
              'etoile': pygame.image.load('Star.png'),
              'coin': pygame.image.load('Wall_Block_Tall.png'),
              'mur': pygame.image.load('Wood_Block_Tall.png'),
              'sol interieur': pygame.image.load('Plain_Block.png'),
              'sol exterieur': pygame.image.load('Grass_Block.png'),
              'titre': pygame.image.load('star_title.png'),
              'resolu': pygame.image.load('star_solved.png'),
              'princesse': pygame.image.load('princess.png'),
              'garcon': pygame.image.load('boy.png'),
              'filleChat': pygame.image.load('catgirl.png'),
              'filleCheval': pygame.image.load('horngirl.png'),
              'filleViolette': pygame.image.load('pinkgirl.png'),
              'rocher': pygame.image.load('Rock.png'),
              'petit arbre': pygame.image.load('Tree_Short.png'),
              'arbre moyen': pygame.image.load('Tree_Tall.png'),
              'arbre ebourifé': pygame.image.load('Tree_Ugly.png')}

# Dictionnaire faisant le lien entre les caractères apparaissant sur la carte
#d'un niveau et les images qui leur correspondent
LiaisonTuile  = {'x': DicoImage['coin'],
               '#': DicoImage['mur'],
               'o': DicoImage['sol interieur'],
               ' ': DicoImage['sol exterieur']}

LiaisonExterieur  = {'1': DicoImage['rocher'],
                      '2': DicoImage['petit arbre'],
                      '3': DicoImage['arbre moyen'],
                      '4': DicoImage['arbre ebourifé']}

#Liste de tous les personnages possibles: le jeu permet de changer de personnage
#à tout moment
# joueurActuel  est un index indiquant le personnahe choisi
joueurActuel  = 0
ImagesJoueur  = [DicoImage['princesse'],
                DicoImage['garcon'],
                DicoImage['filleChat'],
                DicoImage['filleCheval'],
                DicoImage['filleViolette']]

#---------------Partie essai ----------------------------------------
niveaux = lectureFichierNiveaux('Microban0.txt')
indexNiveauEnCours = 0 
niveau = niveaux[indexNiveauEnCours]
carte = modifierCarte(niveau['carte'], niveau['etatInitial']['joueur'])
jeuEtat = copy.deepcopy(niveau['etatInitial'])

surfaceCarte = dessinerScene(carte,jeuEtat,niveau['objectifs'] )
surfaceCarteRect = surfaceCarte.get_rect()
surfaceCarteRect.center = (MILIEU_LARGEUR_ECRAN, MILIEU_HAUTEUR_ECRAN)
ecran.blit(surfaceCarte,surfaceCarteRect)      
pygame.display.update()
game_over = True
while game_over:     

        for event in pygame.event.get(): 
            if event.type == pygame.QUIT:
                game_over = False

pygame.quit()